import { type ReactNode } from "react";
import { type ActorConfig, type HttpAgentOptions } from "@dfinity/agent";
import type { SiwsIdentityContextType } from "./context.type";
import { IDL } from "@dfinity/candid";
import type { SIWS_IDENTITY_SERVICE } from "./service.interface";
/**
 * Re-export types
 */
export * from "./context.type";
export * from "./service.interface";
export * from "./storage.type";
/**
 * React context for managing SIWS (Sign-In with Solana) identity.
 */
export declare const SiwsIdentityContext: import("react").Context<SiwsIdentityContextType | undefined>;
/**
 * Hook to access the SiwsIdentityContext.
 */
export declare const useSiwsIdentity: () => SiwsIdentityContextType;
/**
 * Provider component for the SIWS identity context. Manages identity state and provides authentication-related functionalities.
 *
 * @prop {IDL.InterfaceFactory} idlFactory - Required. The Interface Description Language (IDL) factory for the canister. This factory is used to create an actor interface for the canister.
 * @prop {string} canisterId - Required. The unique identifier of the canister on the Internet Computer network. This ID is used to establish a connection to the canister.
 * @prop {HttpAgentOptions} httpAgentOptions - Optional. Configuration options for the HTTP agent used to communicate with the Internet Computer network.
 * @prop {ActorConfig} actorOptions - Optional. Configuration options for the actor. These options are passed to the actor upon its creation.
 * @prop {ReactNode} children - Required. The child components that the SiwsIdentityProvider will wrap. This allows any child component to access the authentication context provided by the SiwsIdentityProvider.
 *
 * @example
 * ```tsx
 * import { SiwsIdentityProvider } from 'ic-use-siws-identity';
 * import { _SERVICE } from "[DECLARATIONS PATH]/declarations/ic_siws_provider/ic_siws_provider.did";
 * import {
 *   canisterId,
 *   idlFactory,
 * } from "[DECLARATIONS PATH]/declarations/ic_siws_provider/index";
 *
 * ReactDOM.createRoot(document.getElementById("root")!).render(
 *   <React.StrictMode>
 *     <SolanaProviders>
 *       <SiwsIdentityProvider<_SERVICE>
 *         canisterId={canisterId}
 *         idlFactory={idlFactory}
 *       >
 *         <Actors>
 *           <AuthGuard>
 *             <App />
 *           </AuthGuard>
 *         </Actors>
 *       </SiwsIdentityProvider>
 *     </SolanaProviders>
 *   </React.StrictMode>
 * );
 *```
 */
export declare function SiwsIdentityProvider<T extends SIWS_IDENTITY_SERVICE>({ httpAgentOptions, actorOptions, idlFactory, canisterId, children, }: {
    /** Configuration options for the HTTP agent used to communicate with the Internet Computer network. */
    httpAgentOptions?: HttpAgentOptions;
    /** Configuration options for the actor. These options are passed to the actor upon its creation. */
    actorOptions?: ActorConfig;
    /** The Interface Description Language (IDL) factory for the canister. This factory is used to create an actor interface for the canister. */
    idlFactory: IDL.InterfaceFactory;
    /** The unique identifier of the canister on the Internet Computer network. This ID is used to establish a connection to the canister. */
    canisterId: string;
    /** The child components that the SiwsIdentityProvider will wrap. This allows any child component to access the authentication context provided by the SiwsIdentityProvider. */
    children: ReactNode;
}): import("react/jsx-runtime").JSX.Element;
